/*
Copyright (c) 2013, Frank Roberts 
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors may 
		be used to endorse or promote products derived from this software 
		without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL FRANK ROBERTS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/*
	A structure for the file buffer that the CSV came from. This file buffer
	is allocated when a CSV file is read, and is deallocated when a new CSV
	is read into this structure, or when the structure is freed.
*/
typedef struct {
	int size;
	int pos;
	int used;
	char *buf;
} fb_t;

/*
	The table structure is built from a grid of cells. Each cell has
	a right and a down pointer. Usually, only one or the other is used,
	with only the top left cell having something attached to both. The 
	first column is linked from top to bottom, and all rows are linked
	from left to right. 
*/
typedef struct cell cell;
struct cell
{
	fb_t *fb;
	char *contents;
	int csz;
	int cused;
	cell *right;
	cell *down;	
	int rows;
	int columns;
	char escape;
	char delimiter;
	char newline;

};

/*
	A csvtab is a cell*, which means it is a pointer. You should treat it
	like a pointer. I'm not terribly excited about the fact that it doesn't
	_look_ like a pointer. Perhaps I'll change that in the future.

	DO NOT reach into a csvtab and change things. You should use only
	the four methods listed below. You may READ the rows and columns fields,
	but you should NEVER WRITE to any field.

	**If you reach into a csvtab structure and break something, it is NOT
		my fault.
*/
typedef cell *csvtab;

/*
	You should use these four methods and only these four methods to play
	with csvtab structures.
*/

/* 
newTab

Returns an initialized csvtab with the delimiter, escape, and newline
characters set based on user input.

delim: The delimiter character. Conventionally a comma.

escape: The character used for strings that might contain the delimiter.
	Conventionally a double quote.

newline: The end of line character. Conventionally \n.
*/
csvtab newTab(char delim, char esc, char newline);

/*
DEFAULT_TAB

Use this macro to get a csvtab initialized to use the default characters
for the CSV format (according to RFC 4180).
*/
#define DEFAULT_TAB()	newTab(',', '"', '\n')

/*
COLS and ROWS

Use these macros for getting the number of rows and number of columns
in tab.
*/
#define ROWS(tab)	(tab->rows)
#define COLS(tab)	(tab->columns)

/*
	DELIM, ESC, NEWLINE

Use these macros to access the delimiters for a CSV tab.
*/
#define DELIMIT(tab)	(tab->delimiter)
#define ESCAPE(tab)		(tab->escape)
#define NEWLINE(tab)	(tab->newline)

/* 
readTab

Parses a CSV file from in into tab. Sets the rows and columns fields of
	tab. If tab has previously had a CSV read into it but has not been
	freed, the old structures will be freed first.

tab: An initialized csvtab (get this from newTab).

in: A pointer to a FILE structure which contains CSV formatted text.

Exported via csvparse.h 
*/
void readTab(csvtab tab, FILE *in);

/* 
getContent

Returns a pointer to a NULL terminated string. Specifically, the one
at (row,column), where (0,0) is at the top left corner of the CSV, and
the Y-axis is inverted (so positive Y goes down).

tab: A csvtab which has been parsed from a file.

row: The row in the sheet to get.

column: The columnt in the sheet to get.

Exported via csvparse.h 
*/
const char* getContent(csvtab tab, int row, int column);

/*
setDelims

Set the delimiters for tab.

delim: The delimiter character. Conventionally a comma.

escape: The character used for strings that might contain the delimiter.
	Conventionally a double quote.

newline: The end of line character. Conventionally \n. 

Experted in csvparse.h.
*/
void setDelims(csvtab tab, char delim, char escape, char newline);

/*
setCell

Set the contents of the cell at (row,column) in tab. If no cell structure
exists at (row,column), then one is created. If a cell does exist, then the
contents field is replaced.

tab: The csvtab structure we want to modify.

content: The string to put in this cell.

row: The row of the cell we want to modify.

column: The column of the cell we want to modify.

*/
void setCell(csvtab tab, char *content, int row, int column);

/*

writeTab

Write tab to out, formatted according to RFC 4180, except using the 
delimiter characters associated with tab.

tab: The csvtab structure to write out.

out: The stream to write to.

*/
void writeTab(csvtab tab, FILE *out);

/* 
freeTab

Frees all storage associated with tab, except for tab itself.

tab: The csvtab structure to be deallocated.

*/
void freeTab(csvtab tab);
