/*
Copyright (c) 2013, Frank Roberts 
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors may 
		be used to endorse or promote products derived from this software 
		without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL FRANK ROBERTS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* 
I implement parsing for CSVs per RFC 4180. Well, except for line endings.
RFC 4180 specifies DOS line endings, that is, CRLF. I use only LF by
default, however, this parser should handle CRLF just fine.

I believe this library is thread safe.

There is currently no deep-copy mechanism for csvtabs.

*/
#include "csv.h" 

/* Starting sizes for allocating buffers and strings */
#define STARTSZ	128
#define STRSZ	32	

/* Growth factor for growing buffers */
#define GF	1.5
/* They say 1.5 is a better growth factor than 2 */

// define token types (for parser)
#define DELIM	0 
#define NL		1
#define ESC		2
#define ESCx2	3
#define DATA	4
#define END		5
#define ERR		6

/* The parser always wants new cells to inherit the delimiter characters
from the current cells */
#define P_NEWCELL(tab)	(initCell(&(tab->delimiter), &(tab->escape), &(tab->newline)))

/* A token structure for the parser */
typedef struct
{
	int type;
	char value;
} token;

/* Parsing routines */
void parse_FI(csvtab tab, int *rows, int *columns);
void parse_R(csvtab tab, int *columns);
void parse_FE(csvtab tab);
void parse_E(csvtab tab);
void parse_N(csvtab tab);

/* Internal routines used frequently */
void _freeTab(csvtab tab, int rows, int columns);
csvtab initCell(char* delim, char* esc, char* newline);
void appendTofb(fb_t *fb, char c);
csvtab getCell(csvtab tab, int row, int column);

/* 
newTab

Returns an initialized csvtab with the delimiter, escape, and newline
characters set based on user input.

delim: The delimiter character. Conventionally a comma.

escape: The character used for strings that might contain the delimiter.
	Conventionally a double quote.

newline: The end of line character. Conventionally \n.

Exported via csvparse.h 
*/
csvtab newTab(char delim, char escape, char newline)
{
	return initCell(&delim, &escape, &newline);
}

/* 
readTab

Parses a CSV file from in into tab. Sets the rows and columns fields of
	tab. If tab has previously had a CSV read into it but has not been
	freed, the old structures will be freed first.

tab: An initialized csvtab (get this from newTab).

in: A pointer to a FILE structure which contains CSV formatted text.

Exported via csvparse.h 
*/
void readTab(csvtab tab, FILE *in)
{
	/* This tab may have been used before. Try freeing old structures. */
	/* Save the file buffer so we can reuse it. */
	fb_t *save = tab->fb;
	tab->fb = NULL;
	if (tab->down != NULL)
	{_freeTab(tab->down, tab->rows-1, tab->columns);}
	if (tab->right != NULL)
	{_freeTab(tab->right, 1, tab->columns-1);}
	tab->fb = save;

	/* Allocate a new file buffer if necessary. */
	if (tab->fb == NULL)
	{tab->fb = (fb_t*)malloc(sizeof(fb_t));}
	
	tab->fb->pos = 0;
	tab->fb->used = 0;
	/* Assume that if tab->fb->buf is allocated that tab->fb->size is valid */
	if (tab->fb->buf == NULL)
	{
		tab->fb->size = STARTSZ;
		tab->fb->buf = (char *)malloc(sizeof(char) * tab->fb->size);
	}

	/* Read the file */
	int c;
	while ((c = fgetc(in)) != EOF)
	{
		appendTofb(tab->fb, c);
	}
	appendTofb(tab->fb, EOF);
	/* Reset position in buffer for reading */
	tab->fb->pos = 0;

	tab->rows = 0;
	tab->columns = 0;
	/* Parse the CSV from the file buffer */	
	parse_FI(tab, &(tab->rows), &(tab->columns));
}

/* 
getContent

Returns a pointer to a NULL terminated string. Specifically, the one
at (row,column), where (0,0) is at the top left corner of the CSV, and
the Y-axis is inverted (so positive Y goes down).

tab: A csvtab which has been parsed from a file.

row: The row in the sheet to get.

column: The columnt in the sheet to get.

Exported via csvparse.h 
*/
const char* getContent(csvtab tab, int row, int column)
{
	csvtab temp = getCell(tab, row, column);
	if (temp != NULL) {
		return temp->contents;
	} else {
		return "";
	}
}

/* 
freeTab

Frees all storage associated with tab, except for tab itself.

tab: The csvtab structure to be deallocated.

Exported via csvparse.h 
*/
void freeTab(csvtab tab)
{
	_freeTab(tab, tab->rows, tab->columns);
}

/*
A helper function for freeing csvtab structures. Requiring a separate
specification for rows and columns allows for partial freeing of a 
structure.
*/
void _freeTab(csvtab tab, int rows, int columns)
{
	/* Free the file buffer first */
	if (tab->fb != NULL)
	{
		if (tab->fb->buf != NULL)
		{free(tab->fb->buf);}
		free(tab->fb);
		tab->fb = NULL;
	}

	/* Now work through the structure, freeing all of the cells */
	csvtab temp;
	int i, j;
	for (i = rows-1; i >= 0; i--)
	{
		for (j = columns-1; j >= 0; j--)	
		{
			temp = getCell(tab, i, j);
			free(temp);
		}
	}
	tab->right = NULL;
	tab->down = NULL;
}

/*
	Work through the structure until we get to the right cell, then
	return a pointer to it.
*/
csvtab getCell(csvtab tab, int row, int column)
{
	int i = 0;
	while (i < row && tab->down != NULL)
	{
		tab = tab->down;
		i += 1;
	}
	if (tab->down == NULL && i != row)
	{
		return NULL;
	}
	i = 0;
	while (i < column && tab->right != NULL)
	{
		tab = tab->right;
		i += 1;
	}
	if (tab->right == NULL && i != column)
	{
		return NULL;
	}
	return tab;
}

/*
Get a shiny new csvtab structure. All dynamic fields are left unallocated.
By default, use RFC 4180 characters for delimiters, etc.

Note that delim, esc, and newline may be NULL.
*/
csvtab initCell(char* delim, char* esc, char* newline)
{
	csvtab tab = (csvtab)malloc(sizeof(cell));
	tab->fb = NULL;
	tab->contents = NULL;
	tab->csz = 0;
	tab->cused = 0;
	tab->right = NULL;
	tab->down = NULL;
	tab->rows = 0;
	tab->columns = 0;
	/* Default to RFC 4180 characters for these */
	if (esc != NULL) {tab->escape = *esc;}
	else {tab->escape = '"';}
	if (delim != NULL) {tab->delimiter = *delim;}
	else {tab->delimiter = ',';}
	if (newline != NULL) {tab->newline = *newline;}
	else {tab->newline = '\n';}
	return tab;
}

/*
Add a character at the end of a file buffer. Expand the buffer if
necessary.
*/
void appendTofb(fb_t *fb, char c)
{
	if (fb->used == (fb->size - 1))	
	{
		char *temp = (char *)malloc(sizeof(char) * (fb->size * GF));	
		memcpy(temp, fb->buf, fb->size);
		fb->size *= GF;
		free(fb->buf);
		fb->buf = temp;
	}
	fb->buf[fb->used++] = c;
}

/*
Add a character to the string in a cell of the table. Expand the string
if necessary.
*/
void tabApp(char c, csvtab tab)
{
	if (tab->contents == NULL)
	{
		tab->contents = (char *)malloc(sizeof(char) * STRSZ);
		tab->csz = STRSZ;
	}
	if (tab->cused == (tab->csz - 1))
	{
		char *temp = (char *)malloc(sizeof(char) * tab->csz * GF);
		memcpy(temp, tab->contents, tab->csz);
		free(tab->contents);
		tab->contents = temp;
		tab->csz *= GF;
	}
	tab->contents[tab->cused] = c;
	tab->cused++;
}

/*
	Now we start into the parser. I implement the grammar specified in
	RFC 4180, extended to work with different characters for the delimiter,
	the escape character, and the line ending.

	The parser is a recursive descent parser, and it builds a csvtab 
	structure as it goes.
*/

/*
Print the internal parsing symbol associated with typeid to out.
*/
void printType(int typeid, FILE *out)
{
	switch (typeid)
	{
		case DELIM:
			fprintf(out, "DELIM");
			break;
		case NL:
			fprintf(out, "NL");
			break;
		case ESC:
			fprintf(out, "ESC");
			break;
		case ESCx2:
			fprintf(out, "ESCx2");
			break;
		case DATA:
			fprintf(out, "DATA");
			break;
		case END:
			fprintf(out, "END");
			break;
	}
}
/*
Classify the next character. There are only a few special types of
characters, most things are DATA. The only odd one is ESCx2, which is
two escape characters right next to each other. This should translate into
one occurrence of the escape character in the data.
*/
token peek(csvtab tab)
{
	fb_t *fb = tab->fb;
	token rt;
	if (fb->buf[fb->pos] == EOF)
	{
		rt.type = END;
		rt.value = '\0';
	}
	else if (fb->buf[fb->pos] == tab->delimiter)
	{
		rt.type = DELIM;
		rt.value = '\0';
	}
	else if (fb->buf[fb->pos] == tab->newline)
	{
		rt.type = NL;
		rt.value = '\0';
	}
	else if (fb->buf[fb->pos] == tab->escape)
	{
		if (fb->buf[fb->pos+1] == tab->escape)
		{
			rt.type = ESCx2;
			rt.value = fb->buf[fb->pos];
		}
		else
		{
			rt.type = ESC;
			rt.value = '\0';
		}
	}
	else
	{
		rt.type = DATA;
		rt.value = fb->buf[fb->pos];
	}
	return rt;
}

/*
Move our buffer position forward based on the terminal we just saw.
*/
void advance(fb_t *fb, int typeid)
{
	if (typeid == ESCx2)
	{fb->pos += 2;}
	else
	{fb->pos += 1;}
	
}

/* 
If the next character is what we expect, consume it and move on. If not,
print an error.
*/
token match(csvtab tab, int terminal)
{
	token found = peek(tab);
	if (terminal == found.type)
	{advance(tab->fb, terminal);}
	else
	{
		fprintf(stderr, "Could not match token!\n");
		fprintf(stderr, "Expected type: ");
		printType(terminal, stderr);
		fprintf(stderr, "\nFound type: ");
		printType(found.type, stderr);
		fprintf(stderr, "\n");
		exit(0);
	}
	return found;
}

/*

My terminals:
	dq	-> double quote; the escape character
	ddq -> double double quote; two of the escape character
	nl	-> \n; the new line delimiter
	e	-> epsilon, nothing
	c	-> comma; the delimiter
	eof -> end of file character
	txt -> everything else

My Grammar: (From RFC 4180)
	FI -> R
		| RnlFI

	R  -> FE
		| FEcR

	FE -> dqEdq
		| N
		| e

	E  -> txtE
		| cE
		| nlE
		| ddqE
		| e

	N  -> txtN
		| e

First Sets:
	FI:		{dq, txt, c, nl, e}
	R:		{dq, txt, c, e}  
	FE:		{dq, txt, e}
	E:		{txt, c, nl, ddq, e}
	N:		{txt, e}
Follow Sets:
	FI:		{eof}
	R:		{eof, nl}
	FE:		{eof, nl, c}
	E:		{dq, txt, c, nl, ddq}
	N:		{eof, nl, c, txt}	

Good to know I can still do this stuff.
	
*/


void parse_FI(csvtab tab, int *rows, int *columns)
{
	parse_R(tab, columns);
	(*rows)++;
	if (peek(tab).type == NL)
	{
		match(tab, NL);
		if (peek(tab).type != END)
		{
			tab->down = P_NEWCELL(tab);
			tab->down->fb = tab->fb;
			int extra = 0;
			parse_FI(tab->down, rows, &extra);
		}
	}
	if (peek(tab).type == END)
	{match(tab, END);}
}

void parse_R(csvtab tab, int *columns)
{
	parse_FE(tab);
	(*columns)++;
	if (peek(tab).type == DELIM)
	{
		match(tab, DELIM);
		tab->right = P_NEWCELL(tab);
		tab->right->fb = tab->fb;
		parse_R(tab->right, columns);
	}
}

void parse_FE(csvtab tab)
{
	switch (peek(tab).type)
	{
		case ESC:
			match(tab, ESC);
			parse_E(tab);
			tabApp('\0', tab);
			match(tab, ESC);
			break;
		case DATA:
			parse_N(tab);
			tabApp('\0', tab);
			break;
			// otherwise, fall through
	}	
}

void parse_E(csvtab tab)
{
	token t;
	switch(peek(tab).type)
	{
		case DATA:
			t = match(tab, DATA);
			tabApp(t.value, tab);
			parse_E(tab);
			break;
		case DELIM:
			match(tab, DELIM);
			tabApp(tab->delimiter, tab);
			parse_E(tab);
			break;
		case NL:
			match(tab, NL);
			tabApp(tab->newline, tab);
			parse_E(tab);
			break;
		case ESCx2:
			match(tab, ESCx2);
			tabApp(tab->escape, tab);
			tabApp(tab->escape, tab);
			parse_E(tab);
			break;
		// otherwise, fall through
	}
}

void parse_N(csvtab tab)
{
	token t;
	if (peek(tab).type == DATA)
	{
		t = match(tab, DATA);
		tabApp(t.value, tab);
		parse_N(tab);
	}
}

