/*
Copyright (c) 2013, Frank Roberts 
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors may 
		be used to endorse or promote products derived from this software 
		without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL FRANK ROBERTS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include "csv.h"
#include "csvintern.h"

/*
setDelims

Set the delimiters for tab.

delim: The delimiter character. Conventionally a comma.

escape: The character used for strings that might contain the delimiter.
	Conventionally a double quote.

newline: The end of line character. Conventionally \n. 

Experted in csvparse.h.
*/
void setDelims(csvtab tab, char delim, char escape, char newline)
{
	tab->delimiter = delim;
	tab->escape = escape;
	tab->newline = newline;
}

/*
setCell

Set the contents of the cell at (row,column) in tab. If no cell structure
exists at (row,column), then one is created. If a cell does exist, then the
contents field is replaced.

tab: The csvtab structure we want to modify.

content: The string to put in this cell.

row: The row of the cell we want to modify.

column: The column of the cell we want to modify.

*/
void setCell(csvtab tab, char *content, int row, int column)
{
	csvtab temp = tab;
	int i;
	for (i = 0; i < row; i++) {
		if (temp->down == NULL) {
			temp->down = initCell(&(tab->delimiter), 
									&(tab->escape), 
									&(tab->newline));
		}
		temp = temp->down;
	}
	tab->rows = (tab->rows < (i+1)) ? (i+1) : tab->rows;

	for (i = 0; i < column; i++) {
		if (temp->right == NULL) {
			temp->right = initCell(&(tab->delimiter), 
									&(tab->escape), 
									&(tab->newline));
		}
		temp = temp->right;
	}
	tab->columns = (tab->columns < (i+1)) ? (i+1) : tab->columns;
	/* temp should point at the right cell now */

	if (temp->contents != NULL) {
		free(temp->contents);
	} 
	temp->contents = (char *)malloc(sizeof(char)*(strlen(content)+1));
	strncpy(temp->contents, content, strlen(content)+1);
}

/*

writeTab

Write tab to out, formatted according to RFC 4180, except using the 
delimiter characters associated with tab.

tab: The csvtab structure to write out.

out: The stream to write to.

Exported from csvparse.h.
*/
void writeTab(csvtab tab, FILE *out)
{
	int i, j, k;
	const char *next;
	for (i = 0; i < tab->rows; i++)	
	{
		for (j = 0; j < tab->columns; j++) 
		{
			next = getContent(tab, i, j);
			if (next != NULL) 
			{
				fprintf(out, "%c", tab->escape);
				for (k = 0; k < strlen(next); k++) 
				{
					if (next[k] != tab->escape) 
					{fprintf(out, "%c", next[k]);} 
					else 
					{
						fprintf(out, "%c", tab->escape);
						fprintf(out, "%c", tab->escape);
					} // if
				} // for k

				fprintf(out, "%c", tab->escape);
			} // if next != NULL

			/* Don't print a delimiter if we're at the end of a line */
			if (j + 1 != tab->columns) 
			{fprintf(out, "%c", tab->delimiter);}

		} // for j
		fprintf(out, "%c", tab->newline);
	} // for i
}
